
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "rhp.h"
#include "cbp.h"
#include "sif_f.h"
#include "cell.h"
#include "sif.h"

extern void	Error(char *msg),
		RangeList(char *rl,size_t limit,char *target);

#ifdef	_WIN32
#include <process.h>
#include <windows.h>

#define THREAD_T	HANDLE
#define FUNCTION_T	unsigned __stdcall
#define RETURN_VALUE	0

#define INIT_THREAD(T,F,A)	\
	(T=(HANDLE)_beginthreadex(NULL,0,F,(void *)(A),0,NULL))==0
#define TERM_THREAD(T)	\
	WaitForSingleObject(T,INFINITE)==WAIT_FAILED || CloseHandle(T)==0
#else
#include <pthread.h>

#define THREAD_T	pthread_t
#define FUNCTION_T	void *
#define RETURN_VALUE	NULL

#define INIT_THREAD(T,F,A)	pthread_create(&(T),NULL,F,(void *)(A))
#define TERM_THREAD(T)		pthread_join(T,NULL)
#endif

#define INIT_MT(T,F,A)	\
if (INIT_THREAD(T,F,A)) Error(#F " : multi-threading initialization error.")

#define TERM_MT(T,F)	\
	if (TERM_THREAD(T)) Error(#F " : multi-threading termination error.")

#define LEN	2048

static int	N,BPS,Z;
static double	B,S,F1,F2;
static Cell	C;
static FOM	**fom;
static char	path[LEN];

static FUNCTION_T	Store(void *argc)
{
	double	f1,f2;
	int	y,x;
	Cell	**c;
	char	desc[LEN];

	if ((size_t)argc==8) {
	    StoreImageFile_Float(path,N,N,fom,SIF_F_desc);
	    (void)printf("%d\t%s\n",Z,SIF_F_desc);
	}
	else {
	    f1=f2=fom[0][0];
	    for (y=0; y<N; y++)
	    for (x=0; x<N; x++) if (fom[y][x]<f1) f1=fom[y][x]; else
				if (fom[y][x]>f2) f2=fom[y][x];

	    if ((size_t)argc==9) {
		F1=f1; F2=f2; B=f1; S=(f2-f1)/(double)C;
	    }
	    c=(Cell **)fom;
	    for (y=0; y<N; y++)
	    for (x=0; x<N; x++)
		c[y][x]=(fom[y][x]<=F1)?0:
			(fom[y][x]>=F2)?C:(Cell)((fom[y][x]-B)/S+0.5);

	    (void)sprintf(desc,"%lf\t%lf",F1,F2);
	    StoreImageFile(path,N,N,BPS,c,desc);
	    (void)printf("%d\t%lf\t%lf\n",Z,f1,f2);
	}
	return RETURN_VALUE;
}

static double	Log(double d)
{
	return (d>0.0)?log(d):0.0;
}

int	main(int argc,char **argv)
{
	HiPic	hp;
	int	Ox,Oy,L,M;

	if (argc!=4 && argc!=8 && argc!=9 && argc!=11)
	    Error(
"usage : oct_sg2tg HiPic/ Ox Oy {rangeList Dr RA0 {{base step} BPS} TG_format}"
	    );

	InitReadHiPic(argv[1],&hp);

	if (hp.Nt%2) (void)fputs("bad number of views (warning).\n",stderr);

	if ((Ox=atoi(argv[2]))+hp.Nx<=0 || hp.Nx<=Ox ||
	    (Oy=atoi(argv[3]))+hp.Ny<=0 || hp.Ny<=Oy) Error("bad offset.");

	(void)fprintf(stderr,"%d\t%d\t%d\n",N=hp.Nx+abs(Ox),
					    L=hp.Ny-abs(Oy),
					    M=hp.Nt/2);

	if (argc!=4)
{
	char		*target;
	int		l,z,m,y,r,x;
	double		Dr,DO,RA;
	Float		**P,***SG,*sg,**F;
	FOM		*T;
	THREAD_T	t;
	size_t		ac=argc;

	if ((target=(char *)malloc(sizeof(char)*L))==NULL)
	    Error("memory allocation error for range list.");

	RangeList(argv[4],(size_t)L-1,target);

	l=0; for (z=0; z<L; z++) if (target[z]) ++l;

	if (l==0) Error("no slice.");

	Dr=atof(argv[5]);
	DO=(double)(1-N)/2.0;
	RA=atof(argv[6]);

	if (argc!=8) {
	    if ((BPS=atoi(argv[argc-2]))<=0 || BPS>sizeof(Cell)*8)
		Error("bad bits per sample (BPS).");

	    if (argc!=9) {
		B=atof(argv[7]);
		if ((S=atof(argv[8]))<=0.0) Error("bad value step.");
	    }
	    F1=B; F2=B+S*(double)(C=(Cell)((1ULL<<BPS)-1));
	}
	P=InitCBP(N,M);

	if ((SG=(Float ***)malloc(sizeof(Float **)*l))==NULL ||
	    (*SG=(Float **)malloc(sizeof(Float *)*l*M))==NULL ||
	    (**SG=(Float *)malloc(sizeof(Float)*l*M*N))==NULL ||
	    (fom=(FOM **)malloc(sizeof(FOM *)*N))==NULL ||
	    (*fom=(FOM *)malloc(sizeof(FOM)*N*N))==NULL)
	    Error("memory allocation error for sinograms or tomograms.");

	for (m=1; m<M; m++) SG[0][m]=SG[0][m-1]+N;

	for (z=1; z<l; z++) {
	    SG[z]=SG[z-1]+M;

	    for (m=0; m<M; m++) SG[z][m]=SG[z][m-1]+N;
	}
	for (y=1; y<N; y++) fom[y]=fom[y-1]+N;
{
#ifdef	OCT_SBS
	int	r0=N/2,
		r1=r0+N%2;
#else
	int	r0=abs(Ox),
		r1=hp.Nx;
#ifndef	OCT_SA
	int	r2=(Ox<0)?r0-1:r1,
		r3=(Ox<0)?r1:r0-1;
	double	r4=1.0/(double)(r1-r0+1);
#endif
#endif
	int	r00=(Ox<0)?Ox:0,
		r01=(Ox<0)?r0:0,
		r02=(Ox<0)?N:r1,
		r10=(Ox<0)?hp.Nx-1:N-1,
		r11=(Ox<0)?0:r1,
		r12=(Ox<0)?r0:N,
		z0=(Oy<0)?0:Oy,
		z1=(Oy<0)?-Oy:0;

	for (m=0; m<M; m++) {
	    ReadHiPic(&hp,m);
	    for (l=z=0; z<L; z++) if (target[z]) {
		sg=SG[l++][m]; T=hp.T[z+z0]+r00;
		for (r=r01; r<r02; r++) sg[r]=(-Log(T[r]));
	    }
	    ReadHiPic(&hp,m+M);
	    for (l=z=0; z<L; z++) if (target[z]) {
		sg=SG[l++][m]; T=hp.T[z+z1]+r10;
		for (r=r11; r<r12; r++) sg[r]=(-Log(T[-r]));

		for (r=r0; r<r1; r++) sg[r]=
#if	defined(OCT_SBS) || defined(OCT_SA)
		(sg[r]-Log(T[-r]))/2.0;
#else
		((double)abs(r-r2)*sg[r]+(double)abs(r-r3)*(-Log(T[-r])))*r4;
#endif
	    }
	}
}
	for (l=z=0; z<L; z++) if (target[z]) {
	    sg=SG[l][0];
	    for (m=0; m<M; m++)
	    for (r=0; r<N; r++) P[m][r]=(*(sg++));

	    F=CBP(Dr,DO,RA);

	    if (++l!=1) TERM_MT(t,Store);

	    for (y=0; y<N; y++)
	    for (x=0; x<N; x++) fom[y][x]=F[y][x];

	    (void)sprintf(path,argv[argc-1],Z=z); INIT_MT(t,Store,ac);
	}
	TERM_MT(t,Store);

	free(*fom); free(fom); free(**SG); free(*SG); free(SG); TermCBP();

	free(target);
}
	TermReadHiPic(&hp);

	return 0;
}
