
#include <stdio.h>
#include <stdlib.h>
#include "rhp.h"

extern void	Error(char *msg);

#define WORD	unsigned short
#define BYTE	unsigned char

#if	defined(__BYTE_ORDER__) && defined(__ORDER_LITTLE_ENDIAN__)
#define LITTLE_ENDIAN__	(__BYTE_ORDER__==__ORDER_LITTLE_ENDIAN__)
#elif	defined(__linux__)
#include <endian.h>

#define LITTLE_ENDIAN__	(__BYTE_ORDER==__LITTLE_ENDIAN)
#else
#define LITTLE_ENDIAN__	(-1)
#endif

#if	LITTLE_ENDIAN__==1
#define	LittleEndian(pv)	(pv)
#else
static WORD	LittleEndian(WORD pv)
{
#if	LITTLE_ENDIAN__==0
	return (pv<<8)|(pv>>8);
#else
	WORD	w=127;
	BYTE	*b=(BYTE *)&w;

	return (*b==w)?pv:(pv<<8)|(pv>>8);
#endif
}
#endif

static void	Store(char *path,BYTE *header,size_t Nxy,WORD *PV)
{
	FILE	*f=fopen(path,"wb");

	if (f==NULL || fwrite(header,1,64,f)!=64 || fwrite(PV,2,Nxy,f)!=Nxy) {
	    (void)fputs(path,stderr); Error(" : file not stored.");
	}
	(void)fclose(f);
}

#define LEN	2048

int	main(int argc,char **argv)
{
	HiPic	hp;
	double	RC,base,step,x0,x1,x2,t1,t2,t,min=0.0,max=1.0;
	int	Ox,Oy,Nx,Ny,i,l,y,x;
	WORD	*PV,pv;
	size_t	Nxy,n;
	BYTE	header[64];
	char	path[LEN],form[LEN];
	FILE	*file;

	if (argc!=6 && argc!=8)
	    Error("usage : hp_h2f HRCT/ RC Ox Oy {XT_min XT_max} FRCT/");

	InitReadHiPic(argv[1],&hp);

	if ((RC=atof(argv[2])+0.5)<1.0 || RC>(double)(hp.Nx-1))
	    Error("bad position of rotation center.");

	if (abs(Ox=atoi(argv[3]))*2>=hp.Nx)
	    Error("bad value of horizontal offset.");

	Oy=atoi(argv[4]);

	if (argc==8)
	    if ((min=atof(argv[5]))>0.0 ||
		(max=atof(argv[6]))<1.0)
		Error("bad range of X-ray transmittance.");

	(void)printf("%d\t%d\t%d\n",Nx=hp.Nx-abs(Ox),
				    Ny=hp.Ny+abs(Oy),
				    hp.Nt*2);

	if ((PV=(WORD *)malloc(2*(Nxy=Nx*Ny)))==NULL)
	    Error("memory allocation error.");

	header[0]=(BYTE)'I'; header[1]=(BYTE)'M';
	header[2]=header[3]=0;
	header[4]=(BYTE)(Nx&255); header[5]=(BYTE)(Nx>>8);
	header[6]=(BYTE)(Ny&255); header[7]=(BYTE)(Ny>>8);
	header[8]=header[9]=0;
	header[10]=header[11]=0;
	header[12]=2; header[13]=0;
	for (i=14; i<64; i++) header[i]=0;

	base=0.5-min*(step=65535.0/(max-min));

#define XT2PV(XT)	LittleEndian((WORD)(base+step*XT))

	pv=XT2PV(0.0); for (n=0; n<Nxy; n++) PV[n]=pv;

	(void)sprintf(path,"%s/dark.img",argv[argc-1]);
	Store(path,header,Nxy,PV);

	l=0; for (i=hp.Nt*2+2; i>0; i/=10) ++l;

	(void)sprintf(form,"%s/q%%0%dd.img",argv[argc-1],l);

	pv=XT2PV(1.0); for (n=0; n<Nxy; n++) PV[n]=pv;

	(void)sprintf(path,form,1); Store(path,header,Nxy,PV);
	(void)sprintf(path,form,hp.Nt*2+2); Store(path,header,Nxy,PV);

	x0=RC-(double)hp.Nx/2.0+0.5;
	if (Ox<0) {
	    x1=x0-(double)Ox;
	    x2=x0;
	}
	else {
	    x1=x0;
	    x2=x0+(double)Ox;
	}
	for (i=0; i<hp.Nt; i++) {
	    ReadHiPic(&hp,i);

	    n=0; if (i==0) t1=t2=hp.T[0][(int)RC];

	    for (y=0; y<Oy; y++) for (x=0; x<Nx; x++) PV[n++]=pv;

	    for (y=0; y<hp.Ny; y++) {
		for (x=0, x0=x1; x0<0.0; x0+=1.0, x++) PV[n++]=pv;

		for (; x0<(double)hp.Nx && x<Nx; x0+=1.0, x++) {
		    PV[n++]=((t=hp.T[y][(int)x0])<min)?0:(t>max)?65535:XT2PV(t);

		    if (t<t1) t1=t; else if (t>t2) t2=t;
		}
		for (; x<Nx; x++) PV[n++]=pv;
	    }
	    for (y=Oy; y<0; y++) for (x=0; x<Nx; x++) PV[n++]=pv;

	    (void)sprintf(path,form,i+2); Store(path,header,Nxy,PV);

	    for (y=0; y<Oy; y++) for (x=0; x<Nx; x++) PV[--n]=pv;

	    for (y=hp.Ny-1; y>=0; y--) {
		for (x=Nx-1, x0=x2; x0<0.0; x0+=1.0, x--) PV[--n]=pv;

		for (; x0<(double)hp.Nx && x>=0; x0+=1.0, x--) {
		    PV[--n]=((t=hp.T[y][(int)x0])<min)?0:(t>max)?65535:XT2PV(t);

		    if (t<t1) t1=t; else if (t>t2) t2=t;
		}
		for (; x>=0; x--) PV[--n]=pv;
	    }
	    for (y=Oy; y<0; y++) for (x=0; x<Nx; x++) PV[--n]=pv;

	    (void)sprintf(path,form,hp.Nt+i+2); Store(path,header,Nxy,PV);
	}
	(void)printf("%lf\t%lf\n",t1,t2);

	free(PV);

	(void)sprintf(path,"%s/output.log",argv[argc-1]);
	if ((file=fopen(path,"w"))==NULL) {
	    (void)fputs(path,stderr); Error(" : file not open.");
	}
	(void)sprintf(form,"%%0%dd\t%%lf\t%%lf\t%%d\n",l);

	(void)fprintf(file,form,1,hp.OL->c,hp.OL->a,0);

	for (i=0; i<hp.Nt; i++)
	    (void)fprintf(file,form,i+2,hp.OL[i].c,hp.OL[i].a,1);

	for (i=0; i<hp.Nt; i++)
	    (void)fprintf(file,form,hp.Nt+i+2,hp.OL[i].c,hp.OL[i].a+180.0,1);

	(void)fprintf(file,form,hp.Nt+i+2,hp.OL[i-1].c,hp.OL[i-1].a+180.0,0);

	(void)fclose(file);

	TermReadHiPic(&hp);

	return 0;
}
