
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

extern void	Error(char *msg),
		RangeList(char *rl,size_t limit,char *target);

typedef struct {
		int	t,x,y;
		double	d,c;
	} Data;

#define LEN	2048
#define INC	1800
#define EPS	1.0e-7

int	main(int argc,char **argv)
{
	FILE	*F;
	int	N,L,n,t,x,y;
	Data	*D;
	char	S[LEN],*T,
		*mae="memory allocation error.";
	double	d,c,C,T1,T2,X1,X2,TX,Y1,Y2,TY,a,b;

	if (argc!=3 && argc!=4)
	    Error("usage : oct_fit_180 cmp_180.txt rangeList {cmp_180.txt}");

	if (argv[1][0]=='-')
	    F=stdin;
	else
	    if ((F=fopen(argv[1],"r"))==NULL) Error("file not found.");

	N=L=0; D=NULL;
	while (fgets(S,LEN,F)!=NULL)
	    if (((n=sscanf(S,"%d %d %d %lf %lf",&t,&x,&y,&d,&c))==4 || n==5) &&
		t>=0) {
		if (N%INC==0 &&
		    (D=(Data *)realloc(D,sizeof(Data)*(N+INC)))==NULL)
		    Error(mae);

		if ((D[N].t=t)>L) L=t;

		D[N].x=x; D[N].y=y; D[N].d=d; D[N].c=(n==4)?1.0:c; ++N;
	    }

	if (F!=stdin) (void)fclose(F);

	if (N<=2) Error("insufficient data for LSM fitting.");

	if ((D=(Data *)realloc(D,sizeof(Data)*N))==NULL ||
	    (T=(char *)malloc(sizeof(char)*(L+1)))==NULL) Error(mae);

	RangeList(argv[2],(size_t)L,T);

	if (argc==4)
	    if (argv[3][0]=='-')
		F=stdout;
	    else
		if ((F=fopen(argv[3],"w"))==NULL) Error("file not open.");

	C=T1=T2=X1=X2=TX=Y1=Y2=TY=0.0;
	for (n=0; n<N; n++)
	    if (T[t=D[n].t]) {
		C+=(c=D[n].c);
		T1+=(c*(double)t); T2+=(c*(double)t*(double)t);
		X1+=(c*(double)(x=D[n].x)); X2+=(c*(double)x*(double)x);
					    TX+=(c*(double)t*(double)x);
		Y1+=(c*(double)(y=D[n].y)); Y2+=(c*(double)y*(double)y);
					    TY+=(c*(double)t*(double)y);

		if (argc==4)
		    (void)fprintf(F,"%d\t%d\t%d\t%le\t%le\n",t,x,y,D[n].d,c);
	    }
	    else
		if (argc==4) (void)fprintf(F,"%d\n",t);

	if (argc==4 && F!=stdout) (void)fclose(F);

	if (fabs(d=T2*C-T1*T1)<EPS) Error("inadequate data for LSM fitting.");

	a=(C*TX-T1*X1)/d; b=(T2*X1-T1*TX)/d;
	(void)printf("%lf\t%lf\t%lf\n",a,b,sqrt((X2-a*TX-b*X1)/C));
	a=(C*TY-T1*Y1)/d; b=(T2*Y1-T1*TY)/d;
	(void)printf("%lf\t%lf\t%lf\n",a,b,sqrt((Y2-a*TY-b*Y1)/C));

	free(T); free(D); return 0;
}
