
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "rhp.h"

extern void	Error(char *msg),
		RangeList(char *rl,size_t limit,char *target);

static double	Log(double d)
{
	return (d>0.0)?log(d):0.0;
}

#define LEN	2048

int	main(int argc,char **argv)
{
	HiPic	hp;
	int	Ox,Oy,Nx,Ny,Nt;

	if (argc!=4 && argc!=6)
	    Error("usage : oct_sg HiPic/ Ox Oy {rangeList SG_format}");

	InitReadHiPic(argv[1],&hp);

	if (hp.Nt%2) (void)fputs("bad number of views (warning).\n",stderr);

	if ((Ox=atoi(argv[2]))+hp.Nx<=0 || hp.Nx<=Ox ||
	    (Oy=atoi(argv[3]))+hp.Ny<=0 || hp.Ny<=Oy) Error("bad offset.");

	(void)fprintf(stderr,"%d\t%d\t%d\n",Nx=hp.Nx+abs(Ox),
					    Ny=hp.Ny-abs(Oy),
					    Nt=hp.Nt/2);

	if (argc==6)
{
	char	*target,path[LEN];
	int	L,y,t,l,x;
	float	***SG,*sg;
	FOM	*T;
	FILE	*file;
	double	d;

	if ((target=(char *)malloc(sizeof(char)*Ny))==NULL)
	    Error("memory allocation error for range list.");

	RangeList(argv[4],(size_t)Ny-1,target);

	L=0; for (y=0; y<Ny; y++) if (target[y]) ++L;

	if (L==0) Error("no slice.");

	if ((SG=(float ***)malloc(sizeof(float **)*L))==NULL ||
	    (*SG=(float **)malloc(sizeof(float *)*L*Nt))==NULL ||
	    (**SG=(float *)malloc(sizeof(float)*L*Nt*Nx))==NULL)
	    Error("memory allocation error for sinogram.");

	for (t=1; t<Nt; t++) SG[0][t]=SG[0][t-1]+Nx;

	for (l=1; l<L; l++) {
	    SG[l]=SG[l-1]+Nt;

	    for (t=0; t<Nt; t++) SG[l][t]=SG[l][t-1]+Nx;
	}
{
#ifdef	OCT_SBS
	int	x0=Nx/2,
		x1=x0+Nx%2;
#else
	int	x0=abs(Ox),
		x1=hp.Nx;
#ifndef	OCT_SA
	int	x2=(Ox<0)?x0-1:x1,
		x3=(Ox<0)?x1:x0-1;
	double	x4=1.0/(double)(x1-x0+1);
#endif
#endif
	int	x00=(Ox<0)?Ox:0,
		x01=(Ox<0)?x0:0,
		x02=(Ox<0)?Nx:x1,
		x10=(Ox<0)?hp.Nx-1:Nx-1,
		x11=(Ox<0)?0:x1,
		x12=(Ox<0)?x0:Nx,
		y0=(Oy<0)?0:Oy,
		y1=(Oy<0)?-Oy:0;

	for (t=0; t<Nt; t++) {
	    ReadHiPic(&hp,t);
	    for (l=y=0; y<Ny; y++) if (target[y]) {
		sg=SG[l++][t]; T=hp.T[y+y0]+x00;
		for (x=x01; x<x02; x++) sg[x]=(-Log(T[x]));
	    }
	    ReadHiPic(&hp,t+Nt);
	    for (l=y=0; y<Ny; y++) if (target[y]) {
		sg=SG[l++][t]; T=hp.T[y+y1]+x10;
		for (x=x11; x<x12; x++) sg[x]=(-Log(T[-x]));

		for (x=x0; x<x1; x++) sg[x]=
#if	defined(OCT_SBS) || defined(OCT_SA)
		(sg[x]-Log(T[-x]))/2.0;
#else
		((double)abs(x-x2)*sg[x]+(double)abs(x-x3)*(-Log(T[-x])))*x4;
#endif
	    }
	}
}
	if (argv[5][0]=='-' && argv[5][1]=='\0')
	    (void)fprintf(file=stdout,"%d\t%d\t%d\n",Nx,Nt,L);
	else
	    file=NULL;

	for (l=y=0; y<Ny; y++) if (target[y]) {
	    if (file!=stdout) {
		(void)sprintf(path,argv[5],y);
		if ((file=fopen(path,"wb"))==NULL) {
		    (void)fputs(path,stderr); Error(" : file not open.");
		}
		(void)fprintf(file,"%d\t%d\n",Nx,Nt);
	    }
	    sg=SG[l++][0];
	    for (t=0; t<Nt; t++)
	    for (x=0; x<Nx; x++) {
		d=(*(sg++)); (void)fwrite(&d,sizeof(double),1,file);
	    }
	    if (file!=stdout) (void)fclose(file);
	}
	free(**SG); free(*SG); free(SG); free(target);
}
	TermReadHiPic(&hp);

	return 0;
}
