
#include <stdio.h>
#include <stdlib.h>
#include "rhp.h"
#include "sif_f.h"

extern void	Error(char *msg),
		RangeList(char *rl,size_t limit,char *target);

#ifdef	OCT_XP
#include <math.h>

static double	Log(double d)
{
	return (d>0.0)?log(d):0.0;
}

#define COMMAND		"oct_xp"
#define FORMAT		"XP_format"
#define VALUE(f)	(-Log(f))
#else
#define COMMAND		"oct_xt"
#define FORMAT		"XT_format"
#define VALUE(f)	(f)
#endif

#define LEN	2048

int	main(int argc,char **argv)
{
	HiPic	hp;
	int	Ox,Oy,Nx,Ny,Nt;

	if (argc!=4 && argc!=6)
	    Error("usage : " COMMAND " HiPic/ Ox Oy {rangeList " FORMAT "}");

	InitReadHiPic(argv[1],&hp);

	if (hp.Nt%2) (void)fputs("bad number of views (warning).\n",stderr);

	if ((Ox=atoi(argv[2]))+hp.Nx<=0 || hp.Nx<=Ox ||
	    (Oy=atoi(argv[3]))+hp.Ny<=0 || hp.Ny<=Oy) Error("bad offset.");

	(void)fprintf(stderr,"%d\t%d\t%d\n",Nx=hp.Nx+abs(Ox),
					    Ny=hp.Ny-abs(Oy),
					    Nt=hp.Nt/2);

	if (argc==6)
{
	char	*target,path[LEN];
	FOM	**fom,*F,*T;
	int	y,t,x;

	if ((target=(char *)malloc(sizeof(char)*Nt))==NULL ||
	    (fom=(FOM **)malloc(sizeof(FOM *)*Ny))==NULL ||
	    (*fom=(FOM *)malloc(sizeof(FOM)*Ny*Nx))==NULL)
	    Error("memory allocation error.");

	RangeList(argv[4],(size_t)Nt-1,target);

	for (y=1; y<Ny; y++) fom[y]=fom[y-1]+Nx;
{
#ifdef	OCT_SBS
	int	x0=Nx/2,
		x1=x0+Nx%2;
#else
	int	x0=abs(Ox),
		x1=hp.Nx;
#ifndef	OCT_SA
	int	x2=(Ox<0)?x0-1:x1,
		x3=(Ox<0)?x1:x0-1;
	double	x4=1.0/(double)(x1-x0+1);
#endif
#endif
	int	x00=(Ox<0)?Ox:0,
		x01=(Ox<0)?x0:0,
		x02=(Ox<0)?Nx:x1,
		x10=(Ox<0)?hp.Nx-1:Nx-1,
		x11=(Ox<0)?0:x1,
		x12=(Ox<0)?x0:Nx,
		y0=(Oy<0)?0:Oy,
		y1=(Oy<0)?-Oy:0;

	for (t=0; t<Nt; t++) if (target[t]) {
	    ReadHiPic(&hp,t);
	    for (y=0; y<Ny; y++) {
		F=fom[y]; T=hp.T[y+y0]+x00;
		for (x=x01; x<x02; x++) F[x]=VALUE(T[x]);
	    }
	    ReadHiPic(&hp,t+Nt);
	    for (y=0; y<Ny; y++) {
		F=fom[y]; T=hp.T[y+y1]+x10;
		for (x=x11; x<x12; x++) F[x]=VALUE(T[-x]);

		for (x=x0; x<x1; x++) F[x]=
#if	defined(OCT_SBS) || defined(OCT_SA)
		(F[x]+VALUE(T[-x]))/2.0;
#else
		((double)abs(x-x2)*F[x]+(double)abs(x-x3)*VALUE(T[-x]))*x4;
#endif
	    }
	    (void)sprintf(path,argv[5],t);
	    StoreImageFile_Float(path,Nx,Ny,fom,SIF_F_desc);
	    (void)printf("%d\t%s\n",t,SIF_F_desc);
	}
}
	free(*fom); free(fom); free(target);
}
	TermReadHiPic(&hp);

	return 0;
}
