
#include <stdio.h>
#include <stdlib.h>
#include "csi_f.h"
#include "rif_f.h"
#include "sgif.h"

static void	Error(char *msg)
{
	(void)fputs(msg,stderr); (void)fputc('\n',stderr); exit(1);
}

#define BYTE	unsigned char
#define LEN	2048

static void	CM_RGB(char *path,BYTE *R,BYTE *G,BYTE *B)
{
	FILE	*file;
	int	pv,r,g,b,c,pv1,pv2;
	char	map[256],line[LEN],str[LEN],*s,*brl="bad range list.";

	if (path[0]=='-')
	    file=stdin;
	else if ((file=fopen(path,"r"))==NULL) {
	    (void)fputs(path,stderr); Error(" : file not found."); 
	}
	for (pv=0; pv<256; pv++) map[pv]=0;

	while (fgets(line,LEN,file)!=NULL) {
	    switch(sscanf(line,"%s %d %d %d",str,&r,&g,&b)) {
		case EOF: goto Break;
		case 2: b=g=r; break;
		case 4: break;
		default: Error("bad color specification.");
	    }
	    if (r<0 || r>=256 ||
		g<0 || g>=256 ||
		b<0 || b>=256) Error("bad R, G and/or B value.");

	    s=str;

#define GETC()	(*(s++))
#define S2N(n)	for (n=0; c>='0' && c<='9'; c=GETC()) \
		    if ((n=n*10+c-'0')>=256) Error(brl)

	    do {
		while ((c=GETC())==',') ;

		if (c!='\0') {
		    if (c=='-') {
			pv1=0;
			if ((c=GETC())==',' || c=='\0')
			    pv2=255;
			else {
			    S2N(pv2);
			    if (c!=',' && c!='\0') Error(brl);
			}
		    }
		    else {
			S2N(pv1);
			if (c==',' || c=='\0')
			    pv2=pv1;
			else if (c=='-')
			    if ((c=GETC())==',' || c=='\0')
				pv2=255;
			    else {
				S2N(pv2);
				if ((c!=',' && c!='\0') || pv2<pv1) Error(brl);
			    }
			else
			    Error(brl);
		    }
		    for (pv=pv1; pv<=pv2; pv++) {
			map[pv]=1; R[pv]=(BYTE)r; G[pv]=(BYTE)g; B[pv]=(BYTE)b;
		    }
		}
	    } while (c==',');
	}
Break:;
	for (pv=0; pv<256; pv++)
	    if (!map[pv]) Error("insufficient color specification.");

	if (file!=stdin) (void)fclose(file);
}

int	main(int argc,char **argv)
{
	char	**path,*id,str[8][LEN],*u_c,
		*u="using integer PV as is floating-point PV",
		*c="converting PV by TIFF image description tag";
	int	Nx,Ny,Nz,z,X,Y,bps,n,y,x;
	double	Min,Max,B,S,min,max,b,s;
	BYTE	*CM_R,*CM_G,*CM_B,red[256],green[256],blue[256],**byte;
	FILE	*file;
	FOM	**fom,f;

	if (argc<3 || argc>7)
	    Error("usage : t2g_float TIFF/ nameFile {{min max} {CM} GIF}");

	path=CheckSliceImages_Float(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);

	if (argc==6 || argc==7) {
	    Min=atof(argv[3]);
	    if ((Max=atof(argv[4]))<=Min) Error("bad value range.");

	    B=0.5-Min*(S=255.0/(Max-Min));
	}
	if (argc!=3) {
	    if (argc==4 || argc==6)
		CM_R=CM_G=CM_B=StoreGIF_Gray;
	    else 
		CM_RGB(argv[argc-2],CM_R=red,CM_G=green,CM_B=blue);

	    if (argv[argc-1][0]=='-')
		file=stdout;
	    else if ((file=fopen(argv[argc-1],"wb"))==NULL) {
		(void)fputs(argv[argc-1],stderr); Error(" : file not open.");
	    }
	    InitStoreGIF(file,Nx,Ny,CM_R,CM_G,CM_B);
	}
	for (z=0; z<Nz; z++) {
	    if ((bps=ReadImageFile_Float(path[z],&X,&Y,&fom,&id))!=0) {
		if (id!=NULL) {
		    if ((n=sscanf(id,"%s %s %s %s %s %s %s %s",
				     str[0],str[1],str[2],str[3],
				     str[4],str[5],str[6],str[7]))==2 || n==6)
			if (sscanf(str[n-2],"%lf",&min)!=1 ||
			    sscanf(str[n-1],"%lf",&max)!=1 || min>=max)
			    u_c=u;
			else {
			    u_c=c; b=min; s=(max-min)/(double)((1ULL<<bps)-1);
			}
			else
			    u_c=((n!=3 && n!=4 && n!=8) ||
				 sscanf(str[n-2],"%lf",&b)!=1 ||
				 sscanf(str[n-1],"%lf",&s)!=1 || s<=0.0)?u:c;

			if (u_c==c)
			    for (y=0; y<Y; y++)
			    for (x=0; x<X; x++) fom[y][x]=b+s*fom[y][x];

			free(id);
		    }
#ifdef	U_C
		    else
			u_c=u;

		    (void)fprintf(stderr,"%s : %s (warning).\n",path[z],u_c);
#endif
	    }
	    else
		if (id!=NULL) free(id);

	    if (argc!=6 && argc!=7) {
		Min=Max=fom[0][0];
		for (y=0; y<Y; y++)
		for (x=0; x<X; x++)
		    if ((f=fom[y][x])<Min) Min=f; else if (f>Max) Max=f;

		if (argc==3) {
		    (void)printf("%d\t%lf\t%lf\n",z,Min,Max); goto Continue;
		}
		B=0.5-Min*(S=(Min>=Max)?1.0:255.0/(Max-Min));
	    }
	    byte=(BYTE **)fom;
	    for (y=0; y<Y; y++)
	    for (x=0; x<X; x++)
		byte[y][x]=((f=fom[y][x])<Min)?0:(f>Max)?255:(BYTE)(B+S*f);

	    StoreGIF(file,(Nx-X)/2,(Ny-Y)/2,X,Y,NULL,NULL,NULL,byte);
Continue:;
	    for (y=Y-1; y>=0; y--) free(fom[y]); free(fom);
	}
	if (argc!=3) {
	    TermStoreGIF(file); if (file!=stdout) (void)fclose(file);
        }
	for (z=Nz-1; z>=0; z--) free(path[z]); free(path);

	return 0;
}
